#ifndef CYGONCE_HAL_ARCH_H
#define CYGONCE_HAL_ARCH_H

#include <pkgconf/hal.h>
#include <cyg/infra/cyg_type.h>

//-----------------------------------------------------------------------------
// Processor saved states:
// The layout of this structure is also defined in "mpc555.inc", for assembly
// code. Do not change this without changing that (or vice versa).
typedef struct
{
#ifdef CYGDBG_HAL_COMMON_INTERRUPTS_SAVE_MINIMUM_CONTEXT
  cyg_uint32    d[16];			// Integer registers
  double        f[14];			// Floating point registers
#else
  cyg_uint32    d[32];			// Integer registers
  double        f[32];			// Floating point registers
#endif
  cyg_uint32    cr;			// Condition register
  cyg_uint32    xer;			// XER
  cyg_uint32    lr;			// Link register
  cyg_uint32    ctr;			// Count rgister

  // These are saved for exceptions and interrupts, but may also
  // be saved in a context switch if thread-aware debugging is enabled.
  cyg_uint32    msr;			// Machine state register
  cyg_uint32    pc;			// Program counter

  // These are only saved for exceptions and interrupts
  cyg_uint32    vector;                // Vector number

  // These are only saved for exceptions, and are not restored
  // when continued.
  cyg_uint32    hid0;                  // Hw Implementation Dependent
  cyg_uint32    dar;                   // Data Address Reg
  cyg_uint32    dsisr;                 // DSISR
  cyg_uint32    pvr;                   // Processor Version

    // Eventually add BATs, SRs and FP registers too.
} HAL_SavedRegisters;

//-----------------------------------------------------------------------------
// Context information
// The layout of this structure is also defined in "mpc555.inc", for assembly
// code. Do not change this without changing that (or vice versa)
typedef struct
{
#ifdef CYGDBG_HAL_COMMON_CONTEXT_SAVE_MINIMUM
  cyg_uint32    d[18];			// Integer registers
  double        f[18];			// Floating point registers
#else
  cyg_uint32    d[26];			// Integer registers
  double        f[32];			// Floating point registers
#endif
  cyg_uint32    r0;                     // Integer register 0
  cyg_uint32    r1;                     // Integer register 1
  cyg_uint32    r2;                     // Integer register 2
  cyg_uint32    r3;                     // Integer register 3

  cyg_uint32    cr;                     // Condition register
  cyg_uint32    pc;                     // Program counter
  cyg_uint32    msr;                    // Machine state register
  cyg_uint32    lr;                     // Link register

#ifndef CYGDBG_HAL_COMMON_CONTEXT_SAVE_MINIMUM
  cyg_uint32    xer;			// XER
  cyg_uint32    ctr;			// Count register
#endif
} HAL_ThreadContext;

//-----------------------------------------------------------------------------
// Exception handling function.
// This function is defined by the kernel according to this prototype. It is
// invoked from the HAL to deal with any CPU exceptions that the HAL does
// not want to deal with itself. It usually invokes the kernel's exception
// delivery mechanism.
externC void cyg_hal_deliver_exception( CYG_WORD code, CYG_ADDRWORD data );

//-----------------------------------------------------------------------------
// Bit manipulation macros

#define HAL_LSBIT_INDEX(index, mask)    \
    asm ( "neg    11,%1;"               \
          "and    11,11,%1;"            \
          "cntlzw %0,11;"               \
          "subfic %0,%0,31;"            \
          : "=r" (index)                \
          : "r" (mask)                  \
          : "r11"                       \
        );

#define HAL_MSBIT_INDEX(index, mask)                    \
    asm ( "cntlzw %1,%0" : "=r" (mask) : "r" (index) );

//-----------------------------------------------------------------------------
// EABI
#define CYGARC_PPC_STACK_FRAME_SIZE     56      // size of a stack frame

//-----------------------------------------------------------------------------
// Context Initialization
// Initialize the context of a thread.
// Arguments:
// _sparg_ name of variable containing current sp, will be written with new sp
// _thread_ thread object address, passed as argument to entry point
// _entry_ entry point address.
// _id_ bit pattern used in initializing registers, for debugging.

#ifdef CYGDBG_HAL_COMMON_CONTEXT_SAVE_MINIMUM
#define HAL_THREAD_INIT_CONTEXT( _sparg_, _thread_, _entry_, _id_ )              \
    CYG_MACRO_START                                                              \
    register CYG_WORD _sp_ = ((CYG_WORD)_sparg_)-CYGARC_PPC_STACK_FRAME_SIZE;    \
    register HAL_ThreadContext *_context_;                                       \
    int _i_;                                                                     \
    _context_ = (HAL_ThreadContext *)((_sp_) - sizeof(HAL_ThreadContext));       \
    for( _i_ = 0; _i_ < 18; _i_++)                                               \
    {                                                                            \
      (_context_)->d[_i_] = (_id_)|_i_;                                          \
      (_context_)->f[_i_] = (_id_)|_i_;                                          \
    }                                                                            \
    (_context_)->r1 = (CYG_WORD)(_sp_);           /* SP = top of stack      */   \
    (_context_)->r3 = (CYG_WORD)(_thread_);       /* R3 = arg1 = thread ptr */   \
    (_context_)->cr = 0;                          /* CR = 0                 */   \
    (_context_)->pc = (CYG_WORD)(_entry_);        /* set PC for thread dbg  */   \
    (_context_)->lr = (CYG_WORD)(_entry_);        /* LR = entry point       */   \
    _sparg_ = (CYG_ADDRESS)_context_;                                            \
    CYG_MACRO_END
#else
#define HAL_THREAD_INIT_CONTEXT( _sparg_, _thread_, _entry_, _id_ )              \
    CYG_MACRO_START                                                              \
    register CYG_WORD _sp_ = ((CYG_WORD)_sparg_)-CYGARC_PPC_STACK_FRAME_SIZE;    \
    register HAL_ThreadContext *_context_;                                       \
    int _i_;                                                                     \
    _context_ = (HAL_ThreadContext *)((_sp_) - sizeof(HAL_ThreadContext));       \
    for(_i_ = 0; _i_ < 26; _i_++)                                                \
      (_context_)->d[_i_] = (_id_)|_i_;                                          \
    for(_i_ = 0; _i_ < 32; _i_++)                                                \
      (_context_)->f[_i_] = (_id_)|_i_;                                          \
    (_context_)->r1 = (CYG_WORD)(_sp_);           /* SP = top of stack      */   \
    (_context_)->r3 = (CYG_WORD)(_thread_);       /* R3 = arg1 = thread ptr */   \
    (_context_)->cr = 0;                          /* CR = 0                 */   \
    (_context_)->pc = (CYG_WORD)(_entry_);        /* set PC for thread dbg  */   \
    (_context_)->lr = (CYG_WORD)(_entry_);        /* LR = entry point       */   \
    (_context_)->xer = 0;                         /* XER = 0                */   \
    (_context_)->ctr = 0;                         /* CTR = 0                */   \
    _sparg_ = (CYG_ADDRESS)_context_;                                            \
    CYG_MACRO_END
#endif

//-----------------------------------------------------------------------------
// Context switch macros.
// The arguments are pointers to locations where the stack pointer
// of the current thread is to be stored, and from where the sp of the
// next thread is to be fetched.

externC void hal_thread_switch_context( CYG_ADDRESS to, CYG_ADDRESS from );
externC void hal_thread_load_context( CYG_ADDRESS to )
    __attribute__ ((noreturn));

#define HAL_THREAD_SWITCH_CONTEXT(_fspptr_,_tspptr_)                    \
        hal_thread_switch_context((CYG_ADDRESS)_tspptr_,(CYG_ADDRESS)_fspptr_);

#define HAL_THREAD_LOAD_CONTEXT(_tspptr_)                               \
        hal_thread_load_context( (CYG_ADDRESS)_tspptr_ );

//-----------------------------------------------------------------------------
// Execution reorder barrier.
// When optimizing the compiler can reorder code. In multithreaded systems
// where the order of actions is vital, this can sometimes cause problems.
// This macro may be inserted into places where reordering should not happen.

#define HAL_REORDER_BARRIER() asm volatile ( "" : : : "memory" )

//-----------------------------------------------------------------------------
// Breakpoint support
// HAL_BREAKPOINT() is a code sequence that will cause a breakpoint to happen
// if executed.
// HAL_BREAKINST is the value of the breakpoint instruction and 
// HAL_BREAKINST_SIZE is its size in bytes.

#define HAL_BREAKPOINT(_label_)                 \
asm volatile (" .globl  " #_label_ ";"          \
              #_label_":"                       \
              " trap"                           \
    );

#define HAL_BREAKINST           0x7d821008
#define HAL_BREAKINST_SIZE      4

//-----------------------------------------------------------------------------
// Thread register state manipulation for GDB support.
externC void hal_mpc555_get_gdb_regs(void * gdbRegisters, HAL_SavedRegisters * ecosRegisters);
externC void hal_mpc555_set_gdb_regs(HAL_SavedRegisters * ecosRegisters, void * gdbRegisters);

// Translate a stack pointer as saved by the thread context macros above into
// a pointer to a HAL_SavedRegisters structure.
#define HAL_THREAD_GET_SAVED_REGISTERS( _sp_, _regs_ )  \
        (_regs_) = (HAL_SavedRegisters *)(_sp_)

// Copy a set of registers from a HAL_SavedRegisters structure into a
// GDB ordered array.    
/*#define HAL_GET_GDB_REGISTERS( _aregval_, _regs_ )              \
    CYG_MACRO_START                                             \
    CYG_ADDRWORD *_regval_ = (CYG_ADDRWORD *)(_aregval_);       \
    int _i_;                                                    \
                                                                \
    for( _i_ = 0; _i_ < 32; _i_++ )                             \
        _regval_[_i_] = (_regs_)->d[_i_];                       \
                                                                \
    _regval_[64] = (_regs_)->pc;                                \
    _regval_[65] = (_regs_)->msr;                               \
    _regval_[66] = (_regs_)->cr;                                \
    _regval_[67] = (_regs_)->lr;                                \
    _regval_[68] = (_regs_)->ctr;                               \
    _regval_[69] = (_regs_)->xer;                               \
    CYG_MACRO_END*/

// Copy a GDB ordered array into a HAL_SavedRegisters structure.
/*#define HAL_SET_GDB_REGISTERS( _regs_ , _aregval_ )             \
    CYG_MACRO_START                                             \
    CYG_ADDRWORD *_regval_ = (CYG_ADDRWORD *)(_aregval_);       \
    int _i_;                                                    \
                                                                \
    for( _i_ = 0; _i_ < 32; _i_++ )                             \
        (_regs_)->d[_i_] = _regval_[_i_];                       \
                                                                \
    (_regs_)->pc  = _regval_[64];                               \
    (_regs_)->msr = _regval_[65];                               \
    (_regs_)->cr  = _regval_[66];                               \
    (_regs_)->lr  = _regval_[67];                               \
    (_regs_)->ctr = _regval_[68];                               \
    (_regs_)->xer = _regval_[69];                               \
    CYG_MACRO_END*/

#define HAL_GET_GDB_REGISTERS( _aregval_, _regs_)               \
    CYG_MACRO_START                                             \
    hal_mpc555_get_gdb_regs( _aregval_, _regs_);                \
    CYG_MACRO_END

#define HAL_SET_GDB_REGISTERS( _regs_, _aregval_)               \
    CYG_MACRO_START                                             \
    hal_mpc555_set_gdb_regs( _regs_, _aregval_);                \
    CYG_MACRO_END

//-----------------------------------------------------------------------------
// HAL setjmp
// Note: These definitions are repeated in context.S. If changes are required
// remember to update both sets.

#define CYGARC_JMP_BUF_SP        0
#define CYGARC_JMP_BUF_R2        1
#define CYGARC_JMP_BUF_R14       2
#define CYGARC_JMP_BUF_R15       3
#define CYGARC_JMP_BUF_R16       4
#define CYGARC_JMP_BUF_R17       5
#define CYGARC_JMP_BUF_R18       6
#define CYGARC_JMP_BUF_R19       7
#define CYGARC_JMP_BUF_R20       8
#define CYGARC_JMP_BUF_R21       9
#define CYGARC_JMP_BUF_R22      10
#define CYGARC_JMP_BUF_R23      11
#define CYGARC_JMP_BUF_R24      12
#define CYGARC_JMP_BUF_R25      13
#define CYGARC_JMP_BUF_R26      14
#define CYGARC_JMP_BUF_R27      15
#define CYGARC_JMP_BUF_R28      16
#define CYGARC_JMP_BUF_R29      17
#define CYGARC_JMP_BUF_R30      18
#define CYGARC_JMP_BUF_R31      19
#define CYGARC_JMP_BUF_LR       20
#define CYGARC_JMP_BUF_CR       21
#define CYGARC_JMP_BUF_F14	22
#define CYGARC_JMP_BUF_F15	24
#define CYGARC_JMP_BUF_F16	26
#define CYGARC_JMP_BUF_F17	28
#define CYGARC_JMP_BUF_F18	30
#define CYGARC_JMP_BUF_F19	32
#define CYGARC_JMP_BUF_F20	34
#define CYGARC_JMP_BUF_F21	36
#define CYGARC_JMP_BUF_F22	38
#define CYGARC_JMP_BUF_F23	40
#define CYGARC_JMP_BUF_F24	42
#define CYGARC_JMP_BUF_F25	44
#define CYGARC_JMP_BUF_F26	46
#define CYGARC_JMP_BUF_F27	48
#define CYGARC_JMP_BUF_F28	50
#define CYGARC_JMP_BUF_F29	52
#define CYGARC_JMP_BUF_F30	54
#define CYGARC_JMP_BUF_F31	56

#define CYGARC_JMP_BUF_SIZE     58

typedef cyg_uint32 hal_jmp_buf[CYGARC_JMP_BUF_SIZE];

externC int hal_setjmp(hal_jmp_buf env);
externC void hal_longjmp(hal_jmp_buf env, int val);

//-----------------------------------------------------------------------------
// Idle thread code.
// This macro is called in the idle thread loop, and gives the HAL the
// chance to insert code. Typical idle thread behaviour might be to halt the
// processor.

externC void hal_idle_thread_action(cyg_uint32 loop_count);

#define HAL_IDLE_THREAD_ACTION(_count_) hal_idle_thread_action(_count_)

//-----------------------------------------------------------------------------
// Minimal and sensible stack sizes: the intention is that applications
// will use these to provide a stack size in the first instance prior to
// proper analysis.  Idle thread stack should be this big.

//    THESE ARE NOT INTENDED TO BE MICROMETRICALLY ACCURATE FIGURES.
//           THEY ARE HOWEVER ENOUGH TO START PROGRAMMING.
// YOU MUST MAKE YOUR STACKS LARGER IF YOU HAVE LARGE "AUTO" VARIABLES!
 
// This is not a config option because it should not be adjusted except
// under "enough rope" sort of disclaimers.
 
// Stack frame overhead per call. The PPC ABI defines regs 13..31 as callee
// saved. callee saved variables are irrelevant for us as they would contain
// automatic variables, so we only count the caller-saved regs here
// So that makes r0..r12 + cr, xer, lr, ctr:
#define CYGNUM_HAL_STACK_FRAME_SIZE (4 * 17)

// Stack needed for a context switch (ppcref_context_size from ppc.inc) 
#ifdef CYGDBG_HAL_COMMON_CONTEXT_SAVE_MINIMUM
#define CYGNUM_HAL_STACK_CONTEXT_SIZE (26 + (18*2))*4
#else
#define CYGNUM_HAL_STACK_CONTEXT_SIZE (36 + (32*2))*4
#endif

// Interrupt + call to ISR, interrupt_end() and the DSR
#ifdef CYGDBG_HAL_COMMON_INTERRUPTS_SAVE_MINIMUM_CONTEXT
#define CYGNUM_HAL_STACK_INTERRUPT_SIZE \
    (4*(16 + 14*2 + 11) + 2*CYGNUM_HAL_STACK_FRAME_SIZE)
#else
#define CYGNUM_HAL_STACK_INTERRUPT_SIZE \
    (4*(32 + 32*2 + 11) + 2*CYGNUM_HAL_STACK_FRAME_SIZE)
#endif

// We have lots of registers so no particular amount is added in for
// typical local variable usage.

// We define a minimum stack size as the minimum any thread could ever
// legitimately get away with. We can throw asserts if users ask for less
// than this. Allow enough for three interrupt sources - clock, serial and
// one other

#ifdef CYGIMP_HAL_COMMON_INTERRUPTS_USE_INTERRUPT_STACK 

// An interrupt stack which is large enough for all possible interrupt
// conditions (and only used for that purpose) exists.  "User" stacks
// can therefore be much smaller

# define CYGNUM_HAL_STACK_SIZE_MINIMUM \
         (16*CYGNUM_HAL_STACK_FRAME_SIZE + 2*CYGNUM_HAL_STACK_INTERRUPT_SIZE)

#else

// No separate interrupt stack exists.  Make sure all threads contain
// a stack sufficiently large
# define CYGNUM_HAL_STACK_SIZE_MINIMUM                  \
        (((2+3)*CYGNUM_HAL_STACK_INTERRUPT_SIZE) +      \
         (16*CYGNUM_HAL_STACK_FRAME_SIZE))
#endif

// Now make a reasonable choice for a typical thread size. Pluck figures
// from thin air and say 30 call frames with an average of 16 words of
// automatic variables per call frame
#define CYGNUM_HAL_STACK_SIZE_TYPICAL                \
        (CYGNUM_HAL_STACK_SIZE_MINIMUM +             \
         30 * (CYGNUM_HAL_STACK_FRAME_SIZE+(16*4)))

//-----------------------------------------------------------------------------
#endif // CYGONCE_HAL_ARCH_H
