#ifndef CYGONCE_HAL_CMA_SER_H
#define CYGONCE_HAL_CMA_SER_H

//=================================================================================
// Simple serial device driver for the MPC555 based CME555 board.
// Default values are provided. If we want to use something else than the defaults,
// Define some of the macros before including this file
//
// Author : Bob Koninckx
// Date   : 21/07/1999
//
//=================================================================================

//--------------
// Include files
//--------------
#include <pkgconf/hal.h>

#include <cyg/hal/hal_io.h>
#include <cyg/hal/hal_intr.h>
#include <pkgconf/system.h>

#ifdef CYGDBG_HAL_DEBUG_GDB_BREAK_SUPPORT
#include <cyg/hal/mpc555_stub.h>
#endif

//-------------------------------------------------------------
// Unless explicitly asked to export functions, keep them local
// This macro may be defined as e.g. extern before including
// the driver
#ifndef CYG_CMA_PUBLIC
#define CYG_CMA_PUBLIC static
#endif

//-----------------------------------------------------------------------------
// Function names are prefixed with hal_cma_ unless something else is specified
#ifndef CYG_CMA_NAME_PREFIX
#define CYG_CMA_NAME_PREFIX hal_cma_
#endif

// Name conversion macros
#define Str(x, y) x ## y
#define Xstr(x, y) Str(x, y)
#define NC(_name) Xstr(CYG_CMA_NAME_PREFIX, _name)

// The exported functions
CYG_CMA_PUBLIC void NC(init_serial)(void);
CYG_CMA_PUBLIC int  NC(get_char)(void);
CYG_CMA_PUBLIC void NC(put_char)(int c);

//-----------------------------------------------------------------------------
// GDB BREAK Support
// Explicitly define CYG_CMA_STUB before including the driver if support for 
// GDB break handling is desired.
#if defined(CYG_CMA_STUB) && defined(CYGDBG_HAL_DEBUG_GDB_BREAK_SUPPORT)
int  NC(interruptible)(int state);
void NC(init_break_irq)(void);
int cyg_hal_gdb_isr(target_register_t pc); // Called from vectors.s
#endif

//----------------------------------------------------------------------------
// Port to use
// CYG_CMA_PORT must be explicitly defined before including the driver
#if(CYG_CMA_PORT == 1)
#define CYG_DEVICE_SERIAL_RS232_BASE CYG_DEVICE_SERIAL_RS232_BASE_B
#define CYG_DEVICE_SERIAL_INTERRUPT_VECTOR CYGNUM_HAL_INTERRUPT_IMB3_SCI1_RX
#else
#define CYG_DEVICE_SERIAL_RS232_BASE CYG_DEVICE_SERIAL_RS232_BASE_A
#define CYG_DEVICE_SERIAL_INTERRUPT_VECTOR CYGNUM_HAL_INTERRUPT_IMB3_SCI0_RX
#endif

// Controller definitions
#ifndef CYGONCE_HAL_CMA_SER_INL
#define CYGONCE_HAL_CMA_SER_INL

//----------------------------------------------------------------------------
// The CME555 Board implements both serial ports of the MPC555
// Define the base addresses of the SCI control registers on the MPC555
#define CYG_DEVICE_SERIAL_RS232_BASE_A 0x305008 // Base address of SCI0
#define CYG_DEVICE_SERIAL_RS232_BASE_B 0x305020 // Base address of SCI1

//----------------------------------------------------------------------------
// Define CYG_DEVICE_SERIAL_RS232_SCIBR to one of the values below before in-
// cluding the driver if another baudrate than 57600 is desired
// Default baudrate is 38400
// These values are calculated for a 32Mhz clock frequency
#define CYG_DEVICE_SERIAL_RS232_SCIBR_300    4167
#define CYG_DEVICE_SERIAL_RS232_SCIBR_600    2083
#define CYG_DEVICE_SERIAL_RS232_SCIBR_1200   1042
#define CYG_DEVICE_SERIAL_RS232_SCIBR_2400    521
#define CYG_DEVICE_SERIAL_RS232_SCIBR_4800    260
#define CYG_DEVICE_SERIAL_RS232_SCIBR_9600    130 
#define CYG_DEVICE_SERIAL_RS232_SCIBR_14400    65
#define CYG_DEVICE_SERIAL_RS232_SCIBR_19200    43
#define CYG_DEVICE_SERIAL_RS232_SCIBR_28800    38
#define CYG_DEVICE_SERIAL_RS232_SCIBR_38400    33
#define CYG_DEVICE_SERIAL_RS232_SCIBR_57600    22
#define CYG_DEVICE_SERIAL_RS232_SCIBR_115200   11

#ifndef CYG_DEVICE_SERIAL_RS232_SCIBR
#define CYG_DEVICE_SERIAL_RS232_SCIBR CYG_DEVICE_SERIAL_RS232_SCIBR_115200
#endif  

// Define the serial registers
#define CYG_DEVICE_SERIAL_RS232_SCCR0 \
  ((volatile cyg_uint16 *) CYG_DEVICE_SERIAL_RS232_BASE + 0x00)
#define CYG_DEVICE_SERIAL_RS232_SCCR1 \
  ((volatile cyg_uint16 *) CYG_DEVICE_SERIAL_RS232_BASE + 0x01)
#define CYG_DEVICE_SERIAL_RS232_SCSR  \
  ((volatile cyg_uint16 *) CYG_DEVICE_SERIAL_RS232_BASE + 0x02)
#define CYG_DEVICE_SERIAL_RS232_SCDR  \
  ((volatile cyg_uint16 *) CYG_DEVICE_SERIAL_RS232_BASE + 0x03)

// Define the bits in the control registers
#define SCCR0_OTHR   0x8000 // Select baud rate other than system clock
#define SCCR0_LINKBD 0x4000 // Link baud
#define SCCR0_SCxBR  0x1fff // SCI baud rate

#define SCCR1_LOOPS  0x4000 // Loop mode
#define SCCR1_WOMS   0x2000 // Wired or for SCI pins
#define SCCR1_ILT    0x1000 // Idle line detect type
#define SCCR1_PT     0x0800 // Parity type (0 = Odd, 1 = Even)
#define SCCR1_PE     0x0400 // Parity enable
#define SCCR1_M      0x0200 // Mode select (0 = 10 bit frame, 1 = 11 bit frame)
#define SCCR1_WAKE   0x0100 // Wakeup by address mark
#define SCCR1_TIE    0x0080 // Transmit interrupt enable
#define SCCR1_TCIE   0x0040 // Transmit complete interrupt enable
#define SCCR1_RIE    0x0020 // Receiver interrupt enable
#define SCCR1_ILIE   0x0010 // Idle line interrupt enable
#define SCCR1_TE     0x0008 // Transmiter enable
#define SCCR1_RE     0x0004 // Receiver enable
#define SCCR1_RWU    0x0002 // Receiver wakeup
#define SCCR1_SBK    0x0001 // Send break

#define SCSR_TDRE    0x0100 // Transmit data register empty
#define SCSR_TC      0x0080 // Transmit complete
#define SCSR_RDRF    0x0040 // Receive data register full
#define SCSR_RAF     0x0020 // Receive active flag
#define SCSR_IDLE    0x0010 // Idle line detected
#define SCSR_OR      0x0008 // Overrun error
#define SCSR_NF      0x0004 // Noise error flag
#define SCSR_FE      0x0002 // Framing error
#define SCSR_PF      0x0001 // Parity error

#endif // CYGONCE_HAL_CMA_SER_INL

// The minimal init, get and put functions, all by polling
CYG_CMA_PUBLIC void NC(init_serial)(void)
{
  // 8-1-No parity
  HAL_WRITE_UINT16(CYG_DEVICE_SERIAL_RS232_SCCR1, (SCCR1_TE | SCCR1_RE));
  
  // Set baud rate
  HAL_WRITE_UINT16(CYG_DEVICE_SERIAL_RS232_SCCR0, CYG_DEVICE_SERIAL_RS232_SCIBR);
}

CYG_CMA_PUBLIC int NC(get_char)(void)
{
  cyg_uint16 status;
  cyg_uint16 result;

  do {
    HAL_READ_UINT16(CYG_DEVICE_SERIAL_RS232_SCSR, status);
  } while((status & SCSR_RDRF) == 0);

  HAL_READ_UINT16(CYG_DEVICE_SERIAL_RS232_SCDR, result);
  return (unsigned char)(result & 0x00ff);
}

CYG_CMA_PUBLIC void NC(put_char)(int c)
{
  cyg_uint16 status;

  do {
    HAL_READ_UINT16(CYG_DEVICE_SERIAL_RS232_SCSR, status);
  } while((status & SCSR_TDRE) == 0);

  HAL_WRITE_UINT16(CYG_DEVICE_SERIAL_RS232_SCDR, (short)c);

  // Hang around until the character is safely sent
  do {
    HAL_READ_UINT16(CYG_DEVICE_SERIAL_RS232_SCSR, status);
  } while((status & SCSR_TDRE) == 0);
}

//-----------------------------------------------------------
// Additional functions required to provide GDB BREAK support
#if defined(CYG_CMA_STUB) && defined(CYGDBG_HAL_DEBUG_GDB_BREAK_SUPPORT)
int NC(interruptible)(int state)
{
  int vector = CYG_DEVICE_SERIAL_INTERRUPT_VECTOR;
  if(state)
    HAL_INTERRUPT_UNMASK(vector);
  else
    HAL_INTERRUPT_MASK(vector);

  return 0;
}

void NC(init_break_irq)(void)
{
  NC(interruptible)(0);
  HAL_ENABLE_INTERRUPTS();
  NC(interruptible)(1);
}

// This isr is called from the interrupt handler. This should only happen when there is
// no real serial driver, so the code shouldnt mess anything up
int cyg_hal_gdb_isr(target_register_t pc)
{
  // This only works if serial interrupts are configured to happen on the lowest
  // priority
  cyg_uint8 sivec;
  cyg_uint16 scsr;
  cyg_uint16 sccr;
  cyg_uint16 scdr;

  // Find out if the interrupt was caused by the serial port used by the stub
  HAL_READ_UINT8(CYGARC_REG_IMM_SIVEC, sivec);
  if((sivec >> 2) == 0xf)
  {
    HAL_READ_UINT16(CYG_DEVICE_SERIAL_RS232_SCCR1, sccr);
    if(sccr & SCCR1_RIE)
    {
      HAL_READ_UINT16(CYG_DEVICE_SERIAL_RS232_SCSR, scsr);
      if(scsr & SCSR_RDRF)
      {
        HAL_READ_UINT16(CYG_DEVICE_SERIAL_RS232_SCDR, scdr);
        if((unsigned char)(scdr  & 0x00ff) == 3)
        {
          // Ctrl-C: set a breakpoint at PC so GDB will display the
          // correct program context when stopping, rather than the 
          // interrupt handler
          cyg_hal_gdb_interrupt(pc);
  
          // Interrupt handled. Don't call ISR proper. At return
          // from the VSR, execution will stop at the breakpoint
          // just set.
          return 0;
        }
      }
    }
  }

  // Not caused by GDB, Call ISR proper
  return 1;
}

#endif // defined(CYG_CMA_STUB) && defined(CYGDBG_HAL_DEBUG_GDB_BREAK_SUPPORT)

//--------------------------------------
// Undefine all macros used in this file
#undef CYG_CMA_PORT
#undef CYG_CMA_PUBLIC
#undef CYG_CMA_STUB
#undef CYG_CMA_NAME_PREFIX
#undef NC
#undef Str
#undef Str
#undef Xstr
#undef CYG_DEVICE_SERIAL_RS232_BASE

#endif // CYGONCE_HAL_CMA_SER_H
