#ifndef CAN_C
#define CAN_C

//----------------------------------
// Includes and forward declarations
//----------------------------------
#include <cyg/io/io.h>
#include <cyg/io/devtab.h>
#include <cyg/io/can.h>

//--------------------
// Function prototypes
//--------------------
static Cyg_ErrNo can_write(cyg_io_handle_t a_handle, const void * a_buf, cyg_uint32 * a_len);
static Cyg_ErrNo can_read(cyg_io_handle_t a_handle, void * a_buf, cyg_uint32 * a_len);
static Cyg_ErrNo can_get_config(cyg_io_handle_t a_handle, cyg_uint32 a_key, void * a_buf, cyg_uint32 * a_len);
static Cyg_ErrNo can_set_config(cyg_io_handle_t a_handle, cyg_uint32 a_key, const void * a_buf, cyg_uint32 * a_len);

//------------------------------------
// Register the driver with the kernel
//------------------------------------
DEVIO_TABLE(can_devio,
            can_write,
            can_read,
            can_get_config,
            can_set_config);

//-----------------------
// Rhe callback functions
//-----------------------
static void can_init(can_channel * a_chan, bool a_force);
static void can_xmt_message(can_channel * a_chan);
static void can_rcv_message(can_channel * a_chan, const can_message * a_message);
CAN_CALLBACKS(can_callbacks,
              can_init,
              can_xmt_message,
              can_rcv_message);

//--------------------------
// Functions implementations
//--------------------------
// can_init
static void can_init(can_channel * a_chan, bool a_force)
{ 
  int i;

  // Never re-initialise a channel
  if(a_chan->m_init && !a_force)
    return;

  // Initialize the output buffer
  if(a_chan->m_out_buf.m_len)
  {
    a_chan->m_out_buf.m_waiting = false;
    a_chan->m_out_buf.m_abort = false;
    cyg_drv_mutex_init(&a_chan->m_out_buf.m_lock);
    cyg_drv_cond_init(&a_chan->m_out_buf.m_wait, &a_chan->m_out_buf.m_lock);
    a_chan->m_out_buf.m_low_water = a_chan->m_out_buf.m_len / 4;
  }

  // initialize the input buffer
  if(a_chan->m_in_buf.m_len)
  {
    a_chan->m_in_buf.m_waiting = false;
    a_chan->m_in_buf.m_abort = false;
    cyg_drv_mutex_init(&a_chan->m_in_buf.m_lock);
    cyg_drv_cond_init(&a_chan->m_in_buf.m_wait, &a_chan->m_in_buf.m_lock);
  }

  a_chan->m_error = ENOERR;

  if(!a_chan->m_init)
  {
    for(i = 0; i < 256; i++)
      a_chan->m_id_filter[i] = 0;

    for(i = 0; i < 16; i++)
      a_chan->m_rtr_id[i] = 0xFFFF;
  }

  a_chan->m_init = true;
}

// can_read
static Cyg_ErrNo can_read(cyg_io_handle_t a_handle, void * a_buf, cyg_uint32 * a_len)
{
  cyg_devtab_entry_t * t = (cyg_devtab_entry_t *)a_handle;
  can_channel * chan = (can_channel *)t->priv;
  can_message * msg = (can_message *)a_buf;
  can_buf_t * can_buf = &chan->m_in_buf;
  Cyg_ErrNo result = ENOERR;

  bool done = false;
  
  if(*a_len < sizeof(can_message))
    return -EINVAL;

  if(chan->m_error & E_CAN_BOFF)
    return E_CAN_BOFF;

  cyg_drv_mutex_lock(&can_buf->m_lock);
  cyg_drv_dsr_lock(); // avoid races

  while(!done)
  {
    if(can_buf->m_get != can_buf->m_put)
    {
      memcpy(msg, &(can_buf->m_data[can_buf->m_get]), sizeof(can_message));
      if(++can_buf->m_get == can_buf->m_len)
        can_buf->m_get = 0;

      done = true;
    }
    else
    { // Input buffer is empty
      can_buf->m_waiting = true;
      cyg_drv_cond_wait(&can_buf->m_wait);
      if(can_buf->m_abort)
      { // Give up
        can_buf->m_abort = false;
        can_buf->m_waiting = false;
        result = E_CAN_BOFF;

        break;
      }
    }
  }
  
  cyg_drv_dsr_unlock();
  cyg_drv_mutex_unlock(&can_buf->m_lock);

  return result | chan->m_error;
}

// can_write
static Cyg_ErrNo can_write(cyg_io_handle_t a_handle, const void * a_buf, cyg_uint32 * a_len)
{
  cyg_devtab_entry_t * t = (cyg_devtab_entry_t *)a_handle;
  can_channel * chan = (can_channel *)t->priv;
  can_funs * funs = chan->m_funs;
  bool done = false;
  const can_message * msg = (can_message *)a_buf;
  int next;
  can_buf_t * can_buf = &chan->m_out_buf;
  Cyg_ErrNo result = ENOERR;

  if(*a_len != sizeof(can_message))
    return -EINVAL;

  if(chan->m_error & E_CAN_BOFF)
    return E_CAN_BOFF;

  cyg_drv_mutex_lock(&can_buf->m_lock);
  cyg_drv_dsr_lock();  // Avoid race condition testing pointers
  
  while(!done)
  {
    next = can_buf->m_put + 1;
    if(next == can_buf->m_len)
      next = 0;
    if(next == can_buf->m_get)
    {
      can_buf->m_waiting = true;
      // Buffer is full, wait for space
      (funs->start_xmit)(chan);  // Make sure xmit is running, will wait forever otherwise
      cyg_drv_cond_wait(&can_buf->m_wait);
      if(can_buf->m_abort)
      {
        // Give up
        can_buf->m_abort = false;
        can_buf->m_waiting = false;
        result = E_CAN_BOFF;
        
        break;
      }
    }
    else
    {
      memcpy(&(can_buf->m_data[can_buf->m_put]), msg, sizeof(can_message));
      can_buf->m_put = next;

      done = true;
    }
  }

  cyg_drv_dsr_unlock();
  (funs->start_xmit)(chan);

  cyg_drv_mutex_unlock(&can_buf->m_lock);

  return result | chan->m_error;
}

// can_get_config
static Cyg_ErrNo can_get_config(cyg_io_handle_t a_handle, cyg_uint32 a_key, void * a_buf, cyg_uint32 * a_len)
{
  cyg_devtab_entry_t * t = (cyg_devtab_entry_t *)a_handle;
  can_channel * chan = (can_channel *)t->priv;
  Cyg_ErrNo result = ENOERR;

  switch(a_key)
  {
    case CAN_IO_GET_CONFIG_CAN_INFO:
      { // Check the available space
        if(*a_len < sizeof(can_info_t))
          return -EINVAL;

        *(can_info_t *)a_buf = chan->m_config;
        *a_len = sizeof(chan->m_config);
      }
      break;
    default:
      result = -EINVAL;
      break;
  }

  return result;
}

// can_set_config
static Cyg_ErrNo can_set_config(cyg_io_handle_t a_handle, cyg_uint32 a_key, const void * a_buf, cyg_uint32 * a_len)
{
  cyg_devtab_entry_t * t = (cyg_devtab_entry_t *)a_handle;
  can_channel * chan = (can_channel *)t->priv;
  can_funs * funs = chan->m_funs;
  Cyg_ErrNo result = ENOERR;

  switch(a_key)
  { 
    case CAN_IO_ERROR_RESET:
      {
      chan->m_error = ENOERR;
      }
      break;
    case CAN_IO_RESET:
      {
      if(!funs->reset(chan))
        result = -EINVAL;
      }
      break;
    case CAN_IO_SET_CONFIG_CAN_INFO:
      {
      if(*a_len != sizeof(can_info_t))
        return -EINVAL;

      if(!funs->set_config(chan, (can_info_t *)a_buf))
        result = -EINVAL;
      }
      break;
    case CAN_IO_SET_CONFIG_RXID1_INT_ACTION:
      {
      if(*a_len != sizeof(can_rxid_action_t))
        return -EINVAL;

      if(!funs->set_rxid1_action(chan, (can_rxid_action_t *)a_buf))
        result = -EINVAL;
      }
      break;
    case CAN_IO_SET_CONFIG_RXID2_INT_ACTION:
      {
      if(*a_len != sizeof(can_rxid_action_t))
        return -EINVAL;

      if(!funs->set_rxid2_action(chan, (can_rxid_action_t *)a_buf))
        result = -EINVAL;
      }
      break;
    case CAN_IO_ADD_RX_ID:
      {
      if(*a_len != sizeof(cyg_uint16))
        return -EINVAL;

      if(!funs->add_rx_id(chan, *(cyg_uint16 *)a_buf))
        result = -EINVAL;
      }
      break;
    case CAN_IO_REMOVE_RX_ID:
      {
      if(*a_len != sizeof(cyg_uint16))
        return -EINVAL;

      if(!funs->remove_rx_id(chan, *(cyg_uint16 *)a_buf))
        result = -EINVAL;
      }
      break;
    case CAN_IO_ADD_RTR_ID:
      {
      if(*a_len != sizeof(cyg_uint16))
        return -EINVAL;

      if(!funs->add_rtr_id(chan, *(cyg_uint16 *)a_buf))
        result = -EINVAL;
      }
      break;
    case CAN_IO_REMOVE_RTR_ID:
      {
      if(*a_len != sizeof(cyg_uint16))
        return -EINVAL;

      if(!funs->remove_rtr_id(chan, *(cyg_uint16 *)a_buf))
        result = -EINVAL;
      }
      break;
    case CAN_IO_UPDATE_RTR_BUFFER:
      {
      if(*a_len != sizeof(can_message))
        return -EINVAL;
      
      if(!funs->update_rtr_buffer(chan, (can_message *)a_buf))
        result = -EINVAL;
      }
      break;
    default:
      result = -EINVAL;
      break;
  }

  return result;
}

//-------------------
// Callback functions
//-------------------
// can_rcv_message
static void can_rcv_message(can_channel * a_chan, const can_message * a_message)
{
  // Here we got no problem. This will only be called from a dsr and never
  // in a recursive way. So there is absolutely no use in disabling dsr scheduling.
  can_buf_t * can_buf = &a_chan->m_in_buf;

  memcpy(&(can_buf->m_data[can_buf->m_put++]), a_message, sizeof(can_message));
  if(can_buf->m_put == can_buf->m_len)
    can_buf->m_put = 0;

  if(can_buf->m_put == can_buf->m_get)
    a_chan->m_error |= E_CAN_SOFT_OVERRUN;

  if(can_buf->m_waiting)
  {
    can_buf->m_waiting = false;
    cyg_drv_cond_signal(&can_buf->m_wait);
  }
}

// can_xmt_message
static void can_xmt_message(can_channel * a_chan)
{
  can_buf_t * can_buf = &a_chan->m_out_buf;
  can_funs * funs = a_chan->m_funs;
  int space;

  // We 'll use a flag. This may not look so nice. It is surely better than
  // disabling dsr scheduling. It is important that this function never be
  // called when it is running while this can mess up the pointers and cause
  // the complete buffer to be retransmitted again.
  a_chan->m_transmitting = true;

  while(can_buf->m_get != can_buf->m_put)
  {
    if(funs->xmit_message(a_chan, &(can_buf->m_data[can_buf->m_get])))
    {
      can_buf->m_get++;
      if(can_buf->m_get == can_buf->m_len)
        can_buf->m_get = 0;

      if(can_buf->m_waiting)
      {
        space = (can_buf->m_len + can_buf->m_get) - can_buf->m_put;
        if(space > can_buf->m_len)
          space -= can_buf->m_len;
        if(space >= can_buf->m_low_water)
        {
          can_buf->m_waiting = false;
          cyg_drv_cond_broadcast(&can_buf->m_wait);
        }
      }
    }
    else
    {
      a_chan->m_transmitting = false;
      return; // Need to wait for more space
    }
  }

  a_chan->m_transmitting = false;

  if(can_buf->m_waiting)
  {
    can_buf->m_waiting = false;
    cyg_drv_cond_signal(&can_buf->m_wait);
  }
}

#endif // CAN_C
