#ifndef KSYNC_CXX
#define KSYNC_CXX

//----------------------------------
// Includes and forward declarations
//----------------------------------
#include "cyg/kal/ksync.h"
#include "cyg/kal/kdefs.h"
#include <math.h>

//----------------------
// Class implementations
//----------------------
// KSyncObject implementation
// getMark
unsigned int KSyncObject::getMark(unsigned int a_ticks)
{
  return (cyg_current_time() + a_ticks);
}

// KMutex class implementation
// Constructor
KMutex::KMutex()
    : KSyncObject()
{
  cyg_mutex_init(this);
}

// Provide an (empty) virtual destructor
KMutex::~KMutex()
{
  cyg_mutex_destroy(this);
}

// lock
bool KMutex::lock(unsigned int /*a_ticks*/)
{
  if(locked)
    return true;
  else
    return cyg_mutex_lock(this);
}

// unlock
void KMutex::unlock(void)
{
  if(locked)
    cyg_mutex_unlock(this);
}

// KSemaphore class implementation
// Constructor
KSemaphore::KSemaphore(unsigned int a_initialcount)
    : KSyncObject()
{
  cyg_semaphore_init(this, a_initialcount);
}

// Provide an (empty) virtual destructor
KSemaphore::~KSemaphore()
{
  cyg_semaphore_destroy(this);
}

// lock
bool KSemaphore::lock(unsigned int a_ticks)
{
  if(a_ticks)
    return cyg_semaphore_timed_wait(this, getMark(a_ticks));
  else
  {
    cyg_semaphore_wait(this);
    return true;
  }
}

// unlock
void KSemaphore::unlock(void)
{
  cyg_semaphore_post(this);
}

// KCondition class implementation
// Constructor
KCondition::KCondition()
    : KSyncObject()
{
  cyg_cond_init(this, &m_mutex);
}

// Provide an (empty) virtual destructor
KCondition::~KCondition()
{
  cyg_cond_destroy(this);
}

// signal
void KCondition::signal(void)
{
  cyg_cond_broadcast(this);
}

// lock
bool KCondition::lock(unsigned int a_ticks)
{
  if(a_ticks)
    return cyg_cond_timed_wait(this, getMark(a_ticks));
  else
  {
    cyg_cond_wait(this);
    return true;
  }
}

// unlock
void KCondition::unlock(void)
{
}

// KSingleLock implementation
// Constructor
KSingleLock::KSingleLock(KSyncObject * a_syncobject)
    : m_syncobj(a_syncobject)
{
  _ASSERT(a_syncobject != 0, "NULL pointer passed in KSingleLock constructor");
}

// Provide an (empty) virtual destructor
KSingleLock::~KSingleLock()
{
}

// lock
bool KSingleLock::lock(unsigned int a_milisecs)
{
  if(a_milisecs)
    return m_syncobj->lock(convertToTicks(a_milisecs));
  else
    return m_syncobj->lock();
}

// unlock
void KSingleLock::unlock(void)
{
  m_syncobj->unlock();
}

// convertToTicks
unsigned int KSingleLock::convertToTicks(unsigned int a_milisecs)
{
  cyg_handle_t rtc = cyg_real_time_clock();
  cyg_resolution_t res = cyg_clock_get_resolution(rtc);

  float tmp = ((1000000.0 * res.divisor) / res.dividend);
  unsigned int result = (unsigned int)ceil(a_milisecs * tmp);

  return (result ? result : 1);
}

#endif // KSYNC_CXX
