#ifndef KCLOCKTICKGENERATOR_CXX
#define KCLOCKTICKGENERATOR_CXX

//----------------------------------
// Includes and forward declarations
//----------------------------------
#include "kclocktickgenerator.h"
#include "cyg/kal/kinterrupt.h"
#include "cyg/kal/ktimer.h"
#include <cyg/hal/hal_intr.h>
#include <math.h>
#include <pkgconf/kal.h>

//---------------------------
// Some constants definitions
//---------------------------
#define TBREF0 0x2fc204
#define TBREF1 0x2fc208

//-------------------------------
// Some private class definitions
//-------------------------------
// Time base reference A interrupt
class TBRefA_Interrupt : public KInterrupt
{
  public:
                         TBRefA_Interrupt(KHighResolutionClockTickGenerator * a_ctg);
    virtual              ~TBRefA_Interrupt();

  protected:
    // Overloaded virtuals from base class KInterrupt
    virtual KInterrupt::KInterruptStatus isr(void);
    virtual void         dsr(unsigned long a_count);

  private:
    KHighResolutionClockTickGenerator * m_ctg;
    cyg_uint32           m_next_compare;
};

// Time base reference B interrupt
class TBRefB_Interrupt : public KInterrupt
{
  public:
                         TBRefB_Interrupt(KHighResolutionClockTickGenerator * a_ctg);
    virtual              ~TBRefB_Interrupt();

  protected:
    // Overloaded virtuals from base class KInterrupt
    virtual KInterrupt::KInterruptStatus isr(void);
    virtual void         dsr(unsigned long a_count);

  private:
    KHighResolutionClockTickGenerator * m_ctg;
};

//------------------------------
// Private class implementations
//------------------------------
TBRefA_Interrupt::TBRefA_Interrupt(KHighResolutionClockTickGenerator * a_ctg)
    : KInterrupt(CYGNUM_HAL_INTERRUPT_SIU_TB_A,
                 CYGNUM_HAL_INTERRUPT_SIU_TB_A_PRIORITY),
      m_ctg(a_ctg)
{
  // Clear any pending interrupt flags
  acknowledge();

  m_next_compare = KAL_FINE_GRAIN_TIME_PERIOD;
  HAL_WRITE_UINT32(TBREF0, m_next_compare);
  HAL_WRITE_TIMEBASE(0, 0);
  m_next_compare += KAL_FINE_GRAIN_TIME_PERIOD;

  // Enable the interrupts
  unmask();
}

TBRefA_Interrupt::~TBRefA_Interrupt()
{
  // disable the interrupts
  mask();
}

KInterrupt::KInterruptStatus TBRefA_Interrupt::isr(void)
{
  // Do not unmask this interrupt. The ISR runs under Disable interrupt anyway. And if the
  // the interrupt got called again before the DSR was scheduled, it will only be called with
  // a count greater than 1. This is better than risking to get an avalanche of DSR calls that
  // eventually may crash the system.

  HAL_WRITE_UINT32(TBREF0, m_next_compare);

  // First update the TBREFA, only then acknowledge the interrupt. This prevents an avalanche of
  // ISR calls when there is a big ISR latency. We risk to miss a number of ISR (until TBREFB interrupt
  // resets the timebase). That is however better than a crash.

  acknowledge();

  // Cannot do this in the DSR. If for some reason, there is a big latency on the DSR scheduling,
  // the next ISR comes in between 
  // m_next_compare += KAL_FINE_GRAIN_TIME_PERIOD and
  // if(m_next_compare > ...)
  // a compare value that will never be reached will be written to TBREFA, and the interrupt will 
  // remain blocked forever.
  m_next_compare += KAL_FINE_GRAIN_TIME_PERIOD;
  if(m_next_compare > KAL_HIGHRES_NORMALRES_TIME_FACTOR * KAL_FINE_GRAIN_TIME_PERIOD)
    m_next_compare = KAL_FINE_GRAIN_TIME_PERIOD;

  return Partial;
}

void TBRefA_Interrupt::dsr(unsigned long a_count)
{
  m_ctg->generateTicks(a_count);
}

TBRefB_Interrupt::TBRefB_Interrupt(KHighResolutionClockTickGenerator * a_ctg)
    : KInterrupt(CYGNUM_HAL_INTERRUPT_SIU_TB_B,
                 CYGNUM_HAL_INTERRUPT_SIU_TB_B_PRIORITY),
      m_ctg(a_ctg)
{
  // Clear any pending interrupt flags
  acknowledge();

  cyg_uint32 tbrefb = KAL_HIGHRES_NORMALRES_TIME_FACTOR * KAL_FINE_GRAIN_TIME_PERIOD;
  HAL_WRITE_UINT32(TBREF1, tbrefb);
  HAL_WRITE_TIMEBASE(0, 0);

  // Enable the interrupts
  unmask();
}

TBRefB_Interrupt::~TBRefB_Interrupt()
{
  // disable the interrupts
  mask();
}

KInterrupt::KInterruptStatus TBRefB_Interrupt::isr(void)
{
  // Do not unmask this interrupt. The ISR runs under Disable interrupt anyway. And if the
  // the interrupt got called again before the DSR was scheduled, it will only be called with
  // a count greater than 1. This is better than risking to get an avalanche of DSR calls that
  // eventually may crash the system.

  HAL_WRITE_TIMEBASE(0, 0);

  acknowledge();

  return Partial;
}

void TBRefB_Interrupt::dsr(unsigned long a_count)
{
  // interrupt was not masked in the ISR, no need the unmask it here.
  m_ctg->generateTicks(a_count);
}

//------------
// Static Data
//------------
KClockResolution highres = KAL_FINE_GRAIN_TIME_RES;
KClockResolution normalres = {highres.dividend * KAL_HIGHRES_NORMALRES_TIME_FACTOR, highres.divisor};

KClockTickGenerator lowresCTG CYG_INIT_PRIORITY(CLOCK);
KHighResolutionClockTickGenerator normalresCTG(normalres) CYG_INIT_PRIORITY(CLOCK);
KHighResolutionClockTickGenerator highresCTG(highres) CYG_INIT_PRIORITY(CLOCK);

KClockTickGenerator * KClockTickGenerator::m_highres   = &highresCTG;
KClockTickGenerator * KClockTickGenerator::m_normalres = &normalresCTG;
KClockTickGenerator * KClockTickGenerator::m_lowres    = &lowresCTG;

TBRefA_Interrupt tbrefA_int(&highresCTG) CYG_INIT_PRIORITY(APPLICATION);
TBRefB_Interrupt tbrefB_int(&normalresCTG) CYG_INIT_PRIORITY(APPLICATION);

//---------------------
// Class implementation
//---------------------
// Constructor
KClockTickGenerator::KClockTickGenerator()
{
  m_asclock = cyg_real_time_clock();
  cyg_clock_to_counter(m_asclock, &m_ascounter);

  KClockResolution res = cyg_clock_get_resolution(m_asclock);
  m_factor = ((1000.0 * res.divisor) / res.dividend);
}

// Destructor
KClockTickGenerator::~KClockTickGenerator()
{
}

//------------------------------
// Other public member functions
//------------------------------
// timeToTicks
unsigned long KClockTickGenerator::timeToTicks(unsigned long a_microsecs)
{
  return (unsigned int)ceil(a_microsecs * m_factor);
}

// Class KHighResolutionClockTickGenerator
// Constructor
KHighResolutionClockTickGenerator::KHighResolutionClockTickGenerator(KClockResolution a_resolution)
    : KClockTickGenerator()
{
  cyg_clock_create(a_resolution, &m_asclock, this);
  cyg_clock_to_counter(m_asclock, &m_ascounter);

  m_factor = ((1000.0 * a_resolution.divisor) / a_resolution.dividend);
}

// Provide an (empty) virtual destructor
KHighResolutionClockTickGenerator::~KHighResolutionClockTickGenerator()
{
  cyg_clock_delete(m_asclock);
}

// generateTick
void KHighResolutionClockTickGenerator::generateTicks(unsigned long a_ticks)
{
  while(a_ticks)
  {
    cyg_counter_tick(m_ascounter);
    a_ticks--;
  }
}

// currentTime
unsigned long KClockTickGenerator::currentTime(void)
{
  return cyg_counter_current_value(m_ascounter);
}

#endif // KCLOCKTICKGENERATOR_CXX
