//--------------------------------------------------------------
// Motorola S-Record downloader for the CME555 development board
// This is an enhancement to the standard ascii-xfr tool
//--------------------------------------------------------------

//----------------------------------
// Includes and forward declarations
//----------------------------------
#include<stdio.h>
#include<stdlib.h>
#include<unistd.h>
#include<time.h>
#include<string.h>

//----------
// Externals
//----------
extern int getopt();           // Parse options string
extern int optind;             // result from parsing options string
extern char * optarg;          // result from parsing options string

//-------------
// Some globals
//-------------
char * version = "lfstripper 1.00  6-June-1998";

int cdelay = 0;		  // the character delay
int ldelay = 0;		  // the line delay
int verbose = 0;	  // Print verbose information to stderr
time_t start;		  // start time
time_t last;		  // stop time
unsigned long bdone = 0;  // bytes handled
unsigned long btotal = 1; // total bytes to transmit

//-----------------
// Helper functions
//-----------------
// ms_delay
void ms_delay(int a_millisecs)
{ // Put asleep for the requested number of milliseconds
  usleep(1000 * a_millisecs);
}

// lineout
void lineout(char * a_line)
{ // a_line is a null terminated string. Output the line character by character to stdout
  while(*a_line)
  {
    putchar(*a_line++);
    if(cdelay)
    { // Wait if a delay was requested, make sure that the character got out in this case
      fflush(stdout);
      ms_delay(cdelay);
    }
  }
}

// stats
// Show the upload / download statistics on stderr
void stats(int a_force)
{
  if(!verbose)
    return;

  time_t now;
  time_t diff;
  time(&now);
  diff = now-last;
  
  if(!a_force && diff < 2)
    return;
  
  if(diff < 1)
    diff = 1;
  last = now;

  fprintf(stderr, "\r%.1f Kbytes (%.1f %) transferred at %d Bytes per second",
                (float)bdone / 1024, 100.0 * (float)bdone / btotal, (int)(bdone / diff));
  fflush(stderr);
}

// asend
// send a file in ascii mode
int asend(char * a_file)
{
  FILE * fp = fopen(a_file, "r");
  char line[1024];
  char * s;
  int first = 1;

  if(fp == NULL)
  { // report failure if the file cannot be opened for reading
    perror(a_file);
    return -1;
  }

  // Determine the filesize in bytes. It is nice to give the statistics as a percentage done
  unsigned long fsize;
  fseek(fp, 0L, SEEK_END);
  fsize = ftell(fp);
  if(fsize > 0)
    btotal = fsize;
  rewind(fp);

  while(fgets(line, 1024, fp) != NULL)
  {
    if((s = strrchr(line, '\n')) != NULL) 
      *s = 0;

    lineout(line);
    bdone += (strlen(line));

    if(ldelay)
    {
      fflush(stdout);
      ms_delay(ldelay);
    }

    stats(first);
    first = 0;
  }

  fflush(stdout);
  fclose(fp);

  return 0;
}

// usage
// Print usage information to stderr
void usage(void)
{
  fprintf(stderr, "\n%s\n\
Usage: lfstripper [-v] [-h] [-c chardelay] [-l linedelay] inputfilename\n\
         -v : print verbose statistics on standard error\n\
         -h : print this message\n\
          Delays are specified in milliseconds.\n", version);

  exit(1);
}

//--------------
// Main function
//--------------
int main(int argc, char ** argv)
{
  int c;
  char * file;
  int result = -1;

  while((c = getopt(argc, argv, "hvl:c:")) != EOF)
  {
    switch(c)
    {
      case 'v':
        verbose++;
        break;

      case 'l':
        ldelay = atoi(optarg);
        break;

      case 'c':
        cdelay = atoi(optarg);
        break;

      case 'h':
        usage();
        break;

      default:
        fprintf(stderr, "unknown option -%c\n", c); 
        usage();
        break;
    }
  }

  if(optind != argc-1)
  {
    fprintf(stderr, "Missing filename\n");
    usage();
  }

  // OK, everything passed on the command line was correct, we can get to work
  file = argv[optind];

  fprintf(stderr, "ASCII upload of \"%s\"\n", file);
  if(cdelay || ldelay)
    fprintf(stderr, "Line delay: %d ms, Character delay: %d ms\n", ldelay, cdelay);
  fprintf(stderr, "\n");
  fflush(stderr);

  result = asend(file);

  if(verbose)
  {
    stats(1);
    fprintf(stderr, "\r\n... Done.\n");
    fflush(stderr);
  }
 
  return (result < 0 ? 1 : 0);
}
